import warnings

import torch

from .pytorch import SolverModel
from ...graph.concept import EnumConcept


class IMLModel(SolverModel):
    def poi_loss(self, data_item, prop, sensors):
        output_sensor, target_sensor = sensors
        logit = output_sensor(data_item)
        labels = target_sensor(data_item)
        if len(logit) == 0:
            return None

        builder = data_item
        if (builder.needsBatchRootDN()):
            builder.addBatchRootDN()
        datanode = builder.getDataNode(device=self.device)
        concept = prop.sup
        values = []
        try:
            for cdn in datanode.findDatanodes(select=concept):
                value = cdn.getAttribute(f'<{prop.name}>/ILP')
                if isinstance(prop.name, EnumConcept):
                    # if multi-class
                    values.append(value)
                else:
                    values.append(torch.cat((1-value, value), dim=-1))
            if values:
                inference = torch.stack(values)
            else:
                assert logit.shape == (0, 2)
                inference = torch.zeros_like(logit)
        except TypeError:
            message = (f'Failed to get inference result for {prop}. '
                       'Is it included in the inference (with `inference_with` attribute)? '
                       'Continue with predicted value.')
            warnings.warn(message)
            inference = logit.softmax(dim=-1).detach()

        if self.loss:
            local_loss = self.loss[output_sensor, target_sensor](logit, inference, labels)
            return local_loss



from itertools import chain

if __package__ is None or __package__ == '':
    from domiknows.base import Scoped, AutoNamed, NamedTreeNode, NamedTree
else:
    from ..base import Scoped, AutoNamed, NamedTreeNode, NamedTree


@NamedTreeNode.localize_context
class BaseGraphTreeNode(AutoNamed, NamedTreeNode):
    def __init__(self, name=None):
        super().__init__(name)  # name may be update
        super(AutoNamed, self).__init__(self.name)

    def __repr__(self):
        repr_str = f'{type(self).__name__}(name=\'{self.name}\', fullname=\'{self.fullname}\')'
        return repr_str

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    @classmethod
    def clear(cls):
        #AutoNamed.clear() # this call may clear wrong context
        # super() call bring the correct reference to current class `cls`
        super().clear()
        # Sibling class of AutoNamed is NamedTree
        super(AutoNamed, cls).clear()


@BaseGraphTreeNode.share_context
class BaseGraphTree(AutoNamed, NamedTree):
    def __init__(self, name=None):
        super().__init__(name)  # name may be update
        super(AutoNamed, self).__init__(self.name)

    def __repr__(self):
        repr_str = f'{type(self).__name__}(name=\'{self.name}\', fullname=\'{self.fullname}\')'
        return repr_str

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    @classmethod
    def clear(cls):
        # see comment above in BaseGraphTreeNode
        super().clear()
        super(AutoNamed, cls).clear()


@Scoped.class_scope
@BaseGraphTreeNode.share_context
class BaseGraphShallowTree(BaseGraphTree):
    # disable context
    def __enter__(self):
        raise AttributeError(
            '{} object has no attribute __enter__'.format(type(self).__name__))

    def __exit__(self, exc_type, exc_value, traceback):
        raise AttributeError(
            '{} object has no attribute __exit__'.format(type(self).__name__))

    # disable query
    def parse_query_apply(self, func, *names, delim='/', trim=True):
        name, names = self.extract_name(*names, delim=delim, trim=trim)
        if names:
            raise ValueError(('{} cannot have nested elements. Access properties using property name directly.'
                              'Query of names {} is not possibly applied.'.format(type(self).__name__, names)))
        return func(self, name)



from typing import List, Dict, Any, NoReturn
from collections import OrderedDict
import torch
from torch.nn import Module
from allennlp.modules.token_embedders import Embedding, PretrainedBertEmbedder
from allennlp.modules.text_field_embedders import BasicTextFieldEmbedder
from allennlp.nn.util import get_text_field_mask
from ...graph import Property
from ...utils import prod, guess_device
from .base import ReaderSensor, ModuleSensor, SinglePreMaskedSensor, MaskedSensor, PreArgsModuleSensor, SinglePreArgMaskedPairSensor
from .module import Concat, SelfCartesianProduct, SelfCartesianProduct3, NGram, PairTokenDistance, PairTokenDependencyRelation, PairTokenDependencyDistance, LowestCommonAncestor, TripPhraseDistRelation, JointCandidate


class SentenceSensor(ReaderSensor):
    def __init__(
        self,
        reader,
        key: str,
        output_only: bool=False
    ) -> NoReturn:
        ReaderSensor.__init__(self, reader, key, output_only=output_only) # *pres=[]
        self.embedders = OrderedDict() # list of SentenceEmbedderLearner

    def add_embedder(self, key, embedder):
        self.reader.claim(key, embedder)
        self.embedders[key] = embedder

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        # This sensor it self can do nothing
        # mayby with self.embedders something more can happen?
        return None

    @property
    def output_dim(self):
        return ()


class LabelSensor(ReaderSensor):
    def __init__(
        self,
        reader,
        key: str,
        output_only: bool=True
    ) -> NoReturn:
        ReaderSensor.__init__(self, reader, key, output_only=output_only)

    @property
    def output_dim(self):
        return ()

    
class LabelMaskSensor(LabelSensor):
    pass


class ConcatSensor(PreArgsModuleSensor, MaskedSensor):
    def create_module(self):
        return Concat()

    @property
    def output_dim(self):
        output_dim = 0
        for pre_dim in self.pre_dims:
            if len(pre_dim) == 0:
                output_dim += 1
            else:
                output_dim += prod(pre_dim) # assume flatten
        return (output_dim,)

    def get_mask(self, data_item: Dict[str, Any]):
        for pre in self.pres:
            for sensor in pre.find(MaskedSensor):
                return sensor.get_mask(data_item)
            else:
                # not found
                continue
            # found
            break
        else:
            raise RuntimeError('{} require at least one pre-required sensor to be MaskedSensor.'.format(self.fullname))

        return None # not going to here


class CartesianProductSensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return SelfCartesianProduct()

    @property
    def output_dim(self):
        if len(self.pre_dim) == 0:
            output_dim = 2
        else:
            output_dim = prod(self.pre_dim) * 2 # assume flatten
        return (output_dim,)

    def get_mask(self, data_item: Dict[str, Any]):
        for sensor in self.pre.find(MaskedSensor):
            break
        else:
            print(self.pre)
            raise RuntimeError('{} require at least one pre-required sensor to be MaskedSensor.'.format(self.fullname))

        mask = sensor.get_mask(data_item).float()
        ms = mask.size()
        mask = mask.view(ms[0], ms[1], 1).matmul(
            mask.view(ms[0], 1, ms[1]))  # (b,l,l)
        return mask


class CandidateSensor(MaskedSensor):
    pass

class CandidateReaderSensor(ReaderSensor, CandidateSensor):
    pass

class JointCandidateSensor(PreArgsModuleSensor, CandidateSensor):
    def create_module(self):
        return JointCandidate()

    @property
    def output_dim(self):
        return ()

    def get_mask(self, data_item: Dict[str, Any]):
        masks = []
        for pre in self.pres:
            for sensor in pre.find(MaskedSensor):
                break
            else:
                print(self.pre)
                raise RuntimeError('{} require at least one pre-required sensor to be MaskedSensor.'.format(self.fullname))
            # (b,l)
            mask = sensor.get_mask(data_item).float()
            masks.append(mask)

        masks_num = len(masks)
        mask = masks[0]
        for i in range(1, masks_num):
            for j in range(i, masks_num):
                masks[j].unsqueeze_(-2)
            mask = mask.unsqueeze_(-1).matmul(masks[i])
        return mask


class CartesianProduct3Sensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return SelfCartesianProduct3()

    @property
    def output_dim(self):
        if len(self.pre_dim) == 0:
            output_dim = 3
        else:
            output_dim = prod(self.pre_dim) * 3 # assume flatten
        return (output_dim,)

    def get_mask(self, data_item: Dict[str, Any]):
        for sensor in self.pre.find(MaskedSensor):
            break
        else:
            print(self.pre)
            raise RuntimeError('{} require at least one pre-required sensor to be MaskedSensor.'.format(self.fullname))

        mask = sensor.get_mask(data_item).float()
        ms = mask.size()
        #(b,l,l)
        mask1 = mask.view(ms[0], ms[1], 1).matmul(mask.view(ms[0], 1, ms[1]))
        mask2 = mask1.view(ms[0], ms[1], ms[1], 1).matmul(mask.view(ms[0], 1, 1, ms[1]))
        
        return mask2

class SentenceEmbedderSensor(SinglePreMaskedSensor, ModuleSensor):
    def create_module(self):
        self.embedding = Embedding(
            num_embeddings=0, # later load or extend
            embedding_dim=self.embedding_dim,
            pretrained_file=self.pretrained_file,
            vocab_namespace=self.key,
            trainable=False,
        )
        return BasicTextFieldEmbedder({self.key: self.embedding})

    def __init__(
        self,
        key: str,
        embedding_dim: int,
        pre,
        pretrained_file: str=None,
        output_only: bool=False
    ) -> NoReturn:
        self.key = key
        self.embedding_dim = embedding_dim
        self.pretrained_file = pretrained_file
        ModuleSensor.__init__(self, pre, output_only=output_only)

        for pre_sensor in pre.find(SentenceSensor):
            pre_sensor.add_embedder(key, self)
            self.tokens_key = pre_sensor.key # used by reader.update_textfield()
            break
        else:
            raise TypeError()

    def update_context(
        self,
        data_item: Dict[str, Any],
        force=False
    ) -> Dict[str, Any]:
        if self.fullname in data_item and isinstance(data_item[self.fullname], dict):
            data_item[self.fullname + '_index'] = data_item[self.fullname] # reserve
            force = True
        return SinglePreMaskedSensor.update_context(self, data_item, force)

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        return self.module(data_item[self.fullname])

    def get_mask(self, data_item: Dict[str, Any]):
        # TODO: make sure update_context has been called
        return get_text_field_mask(data_item[self.fullname + '_index'])


class SentenceBertEmbedderSensor(SentenceEmbedderSensor):
    def create_module(self):
        self.embedding = PretrainedBertEmbedder(
            pretrained_model = self.pretrained_model,
        )
        return BasicTextFieldEmbedder({self.key: self.embedding},
                                      embedder_to_indexer_map={self.key: [
                                          self.key,
                                          "{}-offsets".format(self.key),
                                          "{}-type-ids".format(self.key)]},
                                      allow_unmatched_keys=True)

    def __init__(
        self,
        key: str,
        pretrained_model: str,
        pre,
        output_only: bool=False
    ) -> NoReturn:
        self.key = key
        self.pretrained_model = pretrained_model
        ModuleSensor.__init__(self, pre, output_only=output_only)

        for pre_sensor in pre.find(SentenceSensor):
            pre_sensor.add_embedder(key, self)
            self.tokens_key = pre_sensor.key # used by reader.update_textfield()
            break
        else:
            raise TypeError()


class NGramSensor(PreArgsModuleSensor, SinglePreMaskedSensor):
    def create_module(self):
        return NGram(self.ngram)

    @property
    def output_dim(self):
        #import pdb; pdb.set_trace()
        return tuple(dim * self.ngram for dim in self.pre_dim)

    def __init__(
        self,
        ngram: int,
        pre: Property,
        output_only: bool=False
    ) -> NoReturn:
        self.ngram = ngram
        PreArgsModuleSensor.__init__(self, pre, output_only=output_only)


class TokenDistantSensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return PairTokenDistance(self.emb_num, self.window)

    def __init__(
        self,
        emb_num: int,
        window: int,
        pre: Property,
        output_only: bool=False
    ) -> NoReturn:
        self.emb_num = emb_num
        self.window = window
        SinglePreArgMaskedPairSensor.__init__(self, pre, output_only=output_only)

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        device, _ = guess_device(data_item).most_common(1)[0]
        self.module.main_module.default_device = device
        return super().forward(data_item)


class TripPhraseDistSensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return TripPhraseDistRelation()

    @property
    def output_dim(self):
        return (self.pre_dim[0] * 2,)

    def get_mask(self, data_item: Dict[str, Any]):
        for sensor in self.pre.find(MaskedSensor):
            break
        else:
            print(self.pre)
            raise RuntimeError('{} require at least one pre-required sensor to be MaskedSensor.'.format(self.fullname))

        mask = sensor.get_mask(data_item).float()
        ms = mask.size()
        #(b,l,l)
        mask1 = mask.view(ms[0], ms[1], 1).matmul(mask.view(ms[0], 1, ms[1]))
        mask2 = mask1.view(ms[0], ms[1], ms[1], 1).matmul(mask.view(ms[0], 1, 1, ms[1]))

        return mask2


class TokenDepSensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return PairTokenDependencyRelation()

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        #import pdb; pdb.set_trace()
        device, _ = guess_device(data_item).most_common(1)[0]
        self.module.main_module.default_device = device
        return super().forward(data_item)


class TokenLcaSensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return LowestCommonAncestor()

    @property
    def output_dim(self):
        return self.pre_dims[1]

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        return PreArgsModuleSensor.forward(self, data_item)


class TokenDepDistSensor(SinglePreArgMaskedPairSensor):
    def create_module(self):
        return PairTokenDependencyDistance(self.emb_num, self.window)

    def __init__(
        self,
        emb_num: int,
        window: int,
        pre: Property,
        output_only: bool=False
    ) -> NoReturn:
        self.emb_num = emb_num
        self.window  = window
        SinglePreArgMaskedPairSensor.__init__(self, pre, output_only=output_only)

    def forward(
        self,
        data_item: Dict[str, Any]
    ) -> Any:
        device, _ = guess_device(data_item).most_common(1)[0]
        self.module.main_module.default_device = device
        return super().forward(data_item)



import warnings

from gurobipy import Model, GRB
from .solver_session import SolverSession


class GurobiSession(SolverSession):
    def __init__(self, silence=True):
        self.model = Model('solver')
        self.silence = silence
        if silence:
            self.model.params.outputflag = 0

    def __str__(self):
        return str(self.model)

    VMAP = {
        SolverSession.VTYPE.BIN: GRB.BINARY,
        SolverSession.VTYPE.INT: GRB.INTEGER,
        SolverSession.VTYPE.DEC: GRB.CONTINUOUS}
    def var(self, vtype, lb, ub, name=None):
        var = self.model.addVar(lb, ub, vtype=self.VMAP[vtype], name=name)
        self.model.update()
        return var

    CMAP = {
        SolverSession.CTYPE.EQ: (1, GRB.EQUAL),
        SolverSession.CTYPE.LT: (-1, GRB.GREATER_EQUAL),
        SolverSession.CTYPE.LE: (1, GRB.LESS_EQUAL),
        SolverSession.CTYPE.GT: (-1, GRB.LESS_EQUAL),
        SolverSession.CTYPE.GE: (1, GRB.GREATER_EQUAL)}
    def constr(self, lhs, ctype, rhs, name=None):
        coeff, ctype = self.CMAP[ctype]
        constr = self.model.addConstr(coeff * lhs, ctype, coeff * rhs, name)
        self.model.update()
        return constr

    OMAP = {
        SolverSession.OTYPE.MAX: GRB.MAXIMIZE,
        SolverSession.OTYPE.MIN: GRB.MINIMIZE}
    def obj(self, otype, expr):
        obj = self.model.setObjective(expr, self.OMAP[otype])
        self.model.update()
        return obj

    def optimize(self):
        self.model.optimize()
        if self.model.status != GRB.Status.OPTIMAL:
            warnings.warn('Model did not finish in an optimal status! Status code is {}.'.format(self.model.status), RuntimeWarning)

    def get_value(self, var):
        return var.x



from domiknows.sensor.pytorch.sensors import JointSensor
from typing import Any
from domiknows.sensor.pytorch.relation_sensors import EdgeSensor


class TokenizerEdgeSensor(EdgeSensor, JointSensor):
    def __init__(self, *pres, relation, edges=None, label=False, device='auto', tokenizer=None):
        super().__init__(*pres, relation=relation, edges=edges, label=label, device=device)
        if not tokenizer:
            raise ValueError('You should select a default Tokenizer')
        self.tokenizer = tokenizer

    def forward(self, text) -> Any:
        tokenized = self.tokenizer.encode_plus(text, return_tensors="pt", return_offsets_mapping=True)
        tokens = tokenized['input_ids'].view(-1).to(device=self.device)
        offset = tokenized['offset_mapping'].to(device=self.device)
        return tokens, offset



from typing import Dict, Any
import torch

from .sensors import FunctionalSensor, FunctionalReaderSensor


class QuerySensor(FunctionalSensor):
    def __init__(self, *pres, **kwargs):
        super().__init__(*pres, **kwargs)
        self.kwinputs = {}

    @property
    def builder(self):
        builder = self.context_helper
        from ...graph import DataNodeBuilder

        if not isinstance(builder, DataNodeBuilder):
            raise TypeError(f'{type(self)} should work with DataNodeBuilder.'
                            'For example, set `build` option to `True` when running the program')
        return builder

    @property
    def concept(self):
        prop = self.sup
        if prop is None:
            raise ValueError('{} must be assigned to property'.format(type(self)))
        concept = prop.sup
        return concept

    def define_inputs(self):
        super().define_inputs()
        datanodes = self.builder.getDataNode(device=self.device).findDatanodes(select=self.concept)
        self.kwinputs['datanodes'] = datanodes

    def forward_wrap(self):
        value = self.forward(*self.inputs, **self.kwinputs)
        if isinstance(value, torch.Tensor) and value.device is not self.device:
            value = value.to(device=self.device)
        return value


class DataNodeSensor(QuerySensor):
    def forward_wrap(self):
        from ...graph import Property
        datanodes = self.kwinputs['datanodes']
        assert len(self.inputs) == len(self.pres)
        inputs = []
        for input, pre in zip(self.inputs, self.pres):
            if isinstance(pre, str):
                try:
                    pre = self.concept[pre]
                except KeyError:
                    pass
            if isinstance(pre, Property) and pre.sup == self.concept:
                assert len(input) == len(datanodes)
                inputs.append(input)
            else:
                # otherwise, repeat the input
                inputs.append([input] * len(datanodes))

        value = [self.forward(*input, datanode=datanode) for datanode, *input in zip(datanodes, *inputs)]

        try:
            return torch.tensor(value, device=self.device)
        except (TypeError, RuntimeError, ValueError):
            return value


class DataNodeReaderSensor(DataNodeSensor, FunctionalReaderSensor):
    pass



from itertools import repeat
from typing import Callable, List
from dataclasses import dataclass

from ..utils import consume, entuple
from .model.base import Mode
from .program import LearningBasedProgram


class ProgramStorageCallback():
    def __init__(self, program, fn) -> None:
        self.program = program
        self.fn = fn
        self.storage = tuple()

    def __call__(self):
        self.storage = self.fn(self.program, *entuple(self.storage))


def hook(callbacks, *args, **kwargs):
    if callbacks:
        consume(callback(*args, **kwargs) for callback in callbacks)


class CallbackProgram(LearningBasedProgram):
    def default_before_train_step(self):
        if self.opt is not None:
            self.opt.zero_grad()

    def default_after_train_step(self, output=None):
        loss, *_ = output
        if self.opt and loss:
            loss.backward()
            self.opt.step()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.before_train = None
        self.after_train = None
        self.before_train_epoch = None
        self.after_train_epoch = None
        self.before_train_step = [self.default_before_train_step]
        self.after_train_step = [self.default_after_train_step]
        self.before_test = None
        self.after_test = None
        self.before_test_epoch = None
        self.after_test_epoch = None
        self.before_test_step = None
        self.after_test_step = None

    def train(self, *args, **kwargs):
        hook(self.before_train)
        super().train(*args, **kwargs)
        hook(self.after_train)

    def train_pure_epoch(self, dataset):
        self.model.mode(Mode.TRAIN)
        self.model.reset()
        for data_item in dataset:
            loss, metric, *output = self.model(data_item)
            yield (loss, metric, *output[:1])

    def train_epoch(self, dataset):
        hook(self.before_train_epoch)
        for _, output in zip(
            map(hook, repeat(self.before_train_step)),
            self.train_pure_epoch(dataset),
            ):
            hook(self.after_train_step, output)
            yield output
        hook(self.after_train_epoch)

    def test(self, *args, **kwargs):
        hook(self.before_test)
        super().test(*args, **kwargs)
        hook(self.after_test)

    def test_epoch(self, dataset):
        hook(self.before_test_epoch)
        for _, output in zip(
            map(hook, repeat(self.before_test_step)),
            super().test_epoch(dataset),
            ):
            hook(self.after_test_step, output)
            yield output
        hook(self.after_test_epoch)



from collections import OrderedDict
from itertools import chain, product
from typing import Type
from .base import Scoped, BaseGraphTree
from ..utils import enum


@Scoped.class_scope
@BaseGraphTree.localize_namespace
class Concept(BaseGraphTree):
    _rels = {}  # catogrory_name : creation callback

    @classmethod
    def relation_type(cls, name=None):
        def update(Rel):
            if name is not None:
                Rel.name = classmethod(lambda cls: name)

            def create(src, *args, auto_constraint=None, **kwargs):
                # add one-by-one
                rels = []
                for argument_name, dst in chain(enum(args, cls=Concept, offset=len(src._out)), enum(kwargs, cls=Concept)):
                    # will be added to _in and _out in Rel constructor
                    rel_inst = Rel(src, dst, argument_name=argument_name, auto_constraint=auto_constraint)
                    rels.append(rel_inst)
                return rels

            cls._rels[Rel.name()] = create
            return Rel

        return update

    def __init__(self, name=None):
        '''
        Declare an concept.
        '''
        BaseGraphTree.__init__(self, name)

        self._in = OrderedDict()  # relation catogrory_name : list of relation inst
        self._out = OrderedDict()  # relation catogrory_name : list of relation inst

    def __str__(self):
        return self.name
    
    def __rept__(self):
        return type(self) + ":" + self.name
    
    def processLCArgs(self, *args, conceptT=None, **kwargs):
        from domiknows.graph.logicalConstrain import eqL, V
        if len(args) > 1 and isinstance(args[1], eqL):
            nameX = args[0]
            path = (nameX, args[1])
                                    
            return [conceptT, V(name=nameX, v=path)]
        elif len(args) and isinstance(args[0], str):
            name = args[0]
            
            if "path" in kwargs:
                path = kwargs['path']
                
                return [conceptT, V(name=name, v=path)]
            else:
                return [conceptT, V(name=name)]
        elif "path" in kwargs:
            path = kwargs['path']
                                    
            return [conceptT, V(name=None, v=path)]
        else:
            return [conceptT]

    def __call__(self, *args, name=None, ConceptClass=None, auto_constraint=None, **kwargs):
        from .relation import IsA, HasA
        if ConceptClass is None:
            ConceptClass = Concept
            
        if (name is None and len(args) and isinstance(args[0], str)) or ("path" in kwargs):
            if isinstance(self, EnumConcept):
                conceptT = (self, self.name, None, len(self.enum))
            else:
                conceptT = (self, self.name, None, 1)

            return self.processLCArgs(*args, conceptT=conceptT, **kwargs)
            
        if (not args and not kwargs) or name is not None:
            new_concept = ConceptClass(name=name, *args, **kwargs)
            new_concept.is_a(self, auto_constraint=auto_constraint)
            return new_concept
        else:
            return self.has_a(*args, auto_constraint=auto_constraint, **kwargs)

    def relate_to(self, concept, *tests):
        from .relation import Relation

        retval = []
        tests_in = [lambda x: x.src == concept,]
        tests_in.extend(tests)
        for rel in chain(*self._in.values()):
            for test in tests_in:
                if isinstance(test, Type) and issubclass(test, Relation):
                    if not isinstance(rel, test):
                        break
                else:
                    if not test(rel):
                        break
            else:
                retval.append(rel)
        tests_out = [lambda x: x.dst == concept,]
        tests_out.extend(tests)
        for rel in chain(*self._out.values()):
            for test in tests_out:
                if isinstance(test, Type) and issubclass(test, Relation):
                    if not isinstance(rel, test):
                        break
                else:
                    if not test(rel):
                        break
            else:
                retval.append(rel)
        return retval

    def __getitem__(self, name):
        try:
            return self.get_sub(name)
        except KeyError as e:
            raise type(e)(name)

    def __setitem__(self, name, obj):
        if isinstance(name, tuple):
            # for name_, obj_ in zip(name, obj):
            #     self[name_] = obj_
            # return self.__setitem__('joint_'+'_'.join(map(str, name)), obj)
            return self.set_apply(name, obj)
        return super().__setitem__(name, obj)

    def set_apply(self, name, sub):
        from ..sensor import Sensor
        from .property import Property
        if isinstance(sub, Property):
            # call usually come from attach, further from constructor of property
            BaseGraphTree.set_apply(self, name, sub)
        elif isinstance(sub, Sensor):
            if name not in self:
                with self:
                    prop = Property(prop_name=name)
            self.get_apply(name).attach(sub)

    def what(self):
        wht = BaseGraphTree.what(self)
        wht['relations'] = dict(self._out)
        return wht

    def __getattr__(self, rel):
        '''
        Create relation by registered relation types
        '''
        cls = type(self)  # bind to the real class

        try:
            Rel = cls._rels[rel]
        except KeyError as e:
            if  isinstance(self, EnumConcept):
                if rel in self.enum:
                    def ecHandle(*args, **kwargs):
                        conceptT = (self, rel, self.get_index(rel), len(self.enum))
                        return self.processLCArgs(*args, conceptT=conceptT, **kwargs)
                    
                    return ecHandle
            else:
                raise AttributeError(*e.args)
        def handle(*args, **kwargs):
            if not args and not kwargs:
                return self._out.setdefault(rel, [])
            return Rel(self, *args, **kwargs)
        return handle

    def get_multiassign(self):
        for prop, value in self.items():
            if len(value) > 1:
                yield self._graph, self, prop, value
                
    def distances(self, p, q):
        '''
        The "distance" used in this concept to measure the consistency.
        Lower value indicates better consistency.
        Feature(s) of one instance is always on only the last axis.
        p, q - [(batch, vdim(s...)),...] * nval
        '''
        # inner product
        #q = self.b.reshape(q, (-1, 1))
        #val = self.b.sum(self.b.matmul(p, q))
        # relative entropy
        #q = self.b.reshape(q, (1, -1))
        #val = self.b.sum(p * self.b.log( p / q ))
        # mse

        nval = len(p)
        assert len(p) == len(q)
        p = self.b.reshape(self.b.concatenete(p, axis=0), (nval, -1, 1))
        q = self.b.reshape(self.b.concatenete(q, axis=0), (1, -1, nval))
        vals = self.b.norm(p - q, axis=1)
        return vals

    def aggregate(self, vals, confs):
        '''
        The aggregation used in this concept to reduce the inconsistent values.
        '''
        vals = self.b.concatenate(vals, axis=0)
        confs = self.b.concatenate(confs, axis=0)
        # TODO: deal with None value in confs. The following is not yet a good solution
        confs[confs is None] = self.b.mean(confs[confs is not None])

        # inverse logistic
        def logit(z): return - self.b.log(self.b(1.) / z - self.b(1.))
        logits = logit(confs)

        # thermodynamic softmax
        def t_softmax(z, beta=-1):  # beta = 1/(k_B*T) - Coldness, the lower, the more concentrated
            z = self.b.exp(- beta * z)
            inf = (z == self.b.inf)
            if self.b.any(inf):
                return inf * 1  # just convert type
            else:
                return z / self.b.sum(z)

        weight = softmax(logits)

        # TODO: should via the same approach as dealing with None values
        vals = self.b.sum(weight * vals)
        confs = self.b.sum(weight * confs)
        return vals, confs

    def bvals(self, prop):
        '''
        Properties: get all binded values

        :param prop: property name
        :type prop: str

        :returns: Return `vals` and `confs` where `vals` is a list of values binded to the property
                  and `confs` is a list of values representing the confidence of each binded value.
                  An element of `vals` should have the shape:
                  ( batch, vdim(s...) )
                  Return `None` is if never binded to this property.
        :rtype: [barray,...], [barray,...]
        '''
        if prop not in self or not self[prop]:
            return [(None, 0), ]
        #vals = []
        #confs = []
        # for val, conf in self[prop]:
        #    vals.append(prop)
        #    confs.append(conf)
        vals, confs = zip(*self[prop])
        return vals, confs

    def rvals(self, prop, hops=1):
        '''
        Properties: get all values from relations
        '''
        vals = []
        confs = []
        for src, rels in self._in.items():
            for rel in rels:
                rvals, rconfs = rel.T(self, prop, hops - 1)
                vals.extend(rvals)
                confs.extend(rconfs)
        return vals, confs

    def vals(self, prop, hops=1):
        vals, confs = self.bvals(prop)
        if hops > 0:
            rvals, rconfs = self.rvals(prop, hops)
            vals.extend(rvals)
            confs.extend(rconfs)
        return vals, confs

    def candidates(self, root_data, query=None, logger = None):
        def basetype(concept):
            # get inheritance rank
            basetypes = []
            for is_a in concept.is_a():
                basetypes.append(basetype(is_a.dst))

            has_a_basetype = []
            for has_a in concept.has_a():
                has_a_basetype.extend(basetype(has_a.dst))
            basetypes.append(has_a_basetype)

            basetypes = sorted(basetypes, key=lambda x: len(x))
            # TODO: here we assume all possible basetypes are homogenous and get the longest one
            # There should be some other semantic to handle this ambiguousity.
            # if still nothing, the original concept is the base type
            return basetypes[-1] or (concept,)

        base = basetype(self)

        assert len(set(base)) == 1  # homogenous base type

        def get_base_data(root_data, single_base):
            from .dataNode import DataNode
            assert isinstance(root_data, DataNode)
            base_data = root_data.findDatanodes(select = single_base.name)
                
            return base_data
        
            while True:
                if base_data[0].getOntologyNode() == single_base:
                    return base_data
                base_data = list(chain(*(bd.getChildDataNodes() for bd in base_data)))

        base_data = get_base_data(root_data, base[0])
        
        if not base_data:
            if logger:
                logger.info('Found base type - %s - for current concept - %s -'%(base[0],self.name))
                conceptNames, relationNames = root_data.findConceptsNamesInDatanodes()
                logger.warning('Found no candidates for - %s -, existing concepts in DataNode are - %s, relations - %s'%(base[0],conceptNames,relationNames))
            
        if query:
            return filter(query, product(base_data, repeat=len(base)))
        else:
            return product(base_data, repeat=len(base))

    def getOntologyGraph(self):  # lowest sub-graph
        node = self
        while isinstance(node, Concept):  # None is not instance of Concept, If this concept has no graph, it will end up None
            node = node.sup
        return node


class EnumConcept(Concept):
    def __init__(self, name=None, values=[]):
        super().__init__(name=name)
        self.enum = values

    @property
    def enum(self):
        return [e.name for e in self._enum]
    
    @property
    def attributes(self):
        return [(self, e.name, self.get_index(e.name), len(self.enum)) for e in self._enum]

    @enum.setter
    def enum(self, values):
        from enum import Enum
        self._enum = Enum(self.name, values, start=0)

    def get_index(self, value):
        return self._enum[value].value

    def get_value(self, index):
        try:
            t = self._enum(index) 
            return t.name
        except ValueError:
            return None
        
    def get_concept(self, value):
        valueIndex = self.get_index(value)
        return (self, value, valueIndex, len(self.enum))
        


if __package__ is None or __package__ == '':
    from domiknows.solver.ilpConfig import ilpConfig 
else:
    from .ilpConfig import ilpConfig 
    
if ilpConfig['ilpSolver'] == "Gurobi":
    from gurobipy import *
elif ilpConfig['ilpSolver'] == "GEKKO":
    from gekko import GEKKO
        
class gekkoILPBooleanProcessor:
    
    # variable controlling what ILP solver is used
    ilpSolver = "Gurobi" # "Gurobi", "GEKKO", "None"
    
    def __init__(self, _ildConfig = ilpConfig) -> None:
        super().__init__()
        
        if _ildConfig is not None:
            self.ilpSolver = _ildConfig['ilpSolver']

    # Negation
    def notVar(m, _var):
        if ilpSolver=="Gurobi":
             _notVar=m.addVar(vtype=GRB.BINARY, name="not_%s" % (_var))
             
             m.addConstr(1 - _var, GRB.EQUAL, _notVar)
        elif ilpSolver == "GEKKO":
            _notVar=m.Var(0, lb=0, ub=1, integer=True, name="not_%s" % (_var))
            
            m.Equation(1 - _var - _notVar == 0)
    
        return _notVar
    
    # Conjunction 2 variable
    def and2Var(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _andVar=m.addVar(vtype=GRB.BINARY, name="and_%s_%s" % (_var1, _var2))
            
            m.addConstr(_andVar - _var1, GRB.LESS_EQUAL, 0)
            m.addConstr(_andVar - _var2, GRB.LESS_EQUAL, 0)
            
            m.addConstr(_var1 + _var2 - _andVar - 1, GRB.LESS_EQUAL, 0)
        elif ilpSolver == "GEKKO":
            _andVar=m.Var(0, lb=0, ub=1, integer=True, name="and_%s_%s" % (_var1, _var2))
    
            m.Equation(_andVar - _var1 <= 0)
            m.Equation(_andVar - _var2 <= 0)
            
            m.Equation(_var1 + _var2 - _andVar <= 1)
            
        return _andVar
    
    # Conjunction
    def andVar(m, *_var):
        if len(_var) < 1:
            return None
        
        if len(_var) == 1:
            for currentVar in _var:
                return currentVar
        
        _andVarName = "and"
        for currentVar in _var:
            _andVarName += "_%s" % (currentVar)
            
        if ilpSolver=="Gurobi":
            _andVar = m.addVar(vtype=GRB.BINARY, name=_andVarName)
        elif ilpSolver == "GEKKO":
            _andVar = m.Var(0, lb=0, ub=1, integer=True, name=_andVarName)
    
        for currentVar in _var:
            if ilpSolver=="Gurobi":
                m.addConstr(_andVar - currentVar, GRB.LESS_EQUAL, 0)
            elif ilpSolver == "GEKKO":
                m.Equation(_andVar - currentVar <= 0)
    
        if ilpSolver=="Gurobi":
            _varSumLinExpr = LinExpr()
            for currentVar in _var:
                _varSumLinExpr.addTerms(1.0, currentVar)
        
            m.addConstr(_varSumLinExpr, GRB.LESS_EQUAL, _andVar + len(_var) - 1)
        elif ilpSolver == "GEKKO":
            m.Equation(m.sum(list(_var)) - _andVar - len(_var) <= -1)
    
        return _andVar
    
    # Disjunction 2 variables
    def or2Var(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _orVar=m.addVar(vtype=GRB.BINARY, name="or_%s_%s" % (_var1, _var2))
            
            m.addConstr(_orVar, GRB.LESS_EQUAL, _var1)
            m.addConstr(_orVar, GRB.LESS_EQUAL, _var2)
            
            m.addConstr(_var1 + _var2, GRB.GREATER_EQUAL, _orVar)
        elif ilpSolver == "GEKKO":
            _orVar=m.Var(0, lb=0, ub=1, integer=True, name="or_%s_%s" % (_var1, _var2))
     
            m.Equation(_orVar - _var1 <= 0)
            m.Equation(_orVar - _var2 <= 0)
            
            m.Equation(_var1 + _var2 - _orVar >= 0)
    
        return _orVar
    
    # Disjunction
    def orVar(m, *_var):
        if len(_var) < 1:
            return None
        
        if len(_var) == 1:
            for currentVar in _var:
                return currentVar
        
        _orVarName = "or"
        for currentVar in _var:
            _orVarName += "_%s" % (currentVar)
        
        if ilpSolver=="Gurobi":
            _orVar = m.addVar(vtype=GRB.BINARY, name=_orVarName)
        elif ilpSolver == "GEKKO":
            _orVar = m.Var(0, lb=0, ub=1, integer=True, name=_orVarName)
    
        for currentVar in _var:
            if ilpSolver=="Gurobi":
                m.addConstr(_orVar, GRB.LESS_EQUAL, currentVar)
            elif ilpSolver == "GEKKO":
                m.Equation(_orVar - currentVar <= 0)
    
        if ilpSolver=="Gurobi":
            _varSumLinExpr = LinExpr()
            for currentVar in _var:
                _varSumLinExpr.addTerms(1.0, currentVar)
        
            m.addConstr(_varSumLinExpr, GRB.GREATER_EQUAL, _orVar)
        elif ilpSolver == "GEKKO":
             m.Equation(m.sum(list(_var)) - _orVar >= 0)
             
        return _orVar
    
    # Nand (Alternative denial) 2 variables
    def nand2Var(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _nandVar = m.addVar(vtype=GRB.BINARY, name="nand_%s_%s" % (_var1, _var2))
            
            m.addConstr(1 - _nandVar, GRB.LESS_EQUAL, _var1)
            m.addConstr(1 - _nandVar, GRB.LESS_EQUAL, _var2)
        
            m.addConstr(_var1 + _var2, GRB.LESS_EQUAL, 2 - _nandVar)
        elif ilpSolver == "GEKKO":
            _nandVar = m.Var(0, lb=0, ub=1, integer=True, name="nand_%s_%s" % (_var1, _var2))
    
            m.Equation(1 - _nandVar - _var1 <= 0)
            m.Equation(1 - _nandVar - _var2 <= 0)
        
            m.Equation(_var1 + _var2 + _nandVar <= 2)
        
        return _nandVar
    
    # Nand (Alternative denial)
    def nandVar(m, *_var):
        if len(_var) < 1:
            return None
        
        if len(_var) == 1:
            for currentVar in _var:
                return currentVar
        
        _nandVarName = "nand"
        for currentVar in _var:
            _nandVarName += "_%s" % (currentVar)
            
        if ilpSolver=="Gurobi":
            _nandVar = m.addVar(vtype=GRB.BINARY, name=_nandVarName)
        elif ilpSolver == "GEKKO":
            _nandVar = m.Var(0, lb=0, ub=1, integer=True, name=_nandVarName)
    
        for currentVar in _var:
            if ilpSolver=="Gurobi":
                m.addConstr(1 - _nandVar, GRB.LESS_EQUAL, currentVar)
            elif ilpSolver == "GEKKO":
                m.Equation(currentVar + _nandVar >= 1)
        
        if ilpSolver=="Gurobi":
            _varSumLinExpr = LinExpr()
            for currentVar in _var:
                _varSumLinExpr.addTerms(1.0, currentVar)
    
            m.addConstr(_varSumLinExpr, GRB.LESS_EQUAL, len(_var) - _nandVar)
        elif ilpSolver == "GEKKO":
            m.Equation(m.sum(list(_var)) + _nandVar <= len(_var))
    
        return _nandVar
    
    # Nor (Joint Denial) i2 variables
    def nor2Var(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _norVar = m.addVar(vtype=GRB.BINARY, name="nor_%s_%s"%(_var1, _var2))
            
            m.addConstr(_var1, GRB.LESS_EQUAL, 1 - _norVar)
            m.addConstr(_var2, GRB.LESS_EQUAL, 1 - _norVar)
            
            m.addConstr(_var1 + _var2, GRB.GREATER_EQUAL, 1 - _norVar)
        elif ilpSolver == "GEKKO":
            _norVar = m.Var(0, lb=0, ub=1, integer=True, name="nor_%s_%s"%(_var1, _var2))
            
            m.Equation(_norVar + _var1 <= 1)
            m.Equation(_norVar + _var2 <= 1)
            
            m.Equation(_var1 + _var2 + _norVar >= 1)
            
        return _norVar
    
    # Nor (Joint Denial)
    def norVar(m, *_var):
        if len(_var) < 1:
            return None
        
        if len(_var) == 1:
            for currentVar in _var:
                return currentVar
        
        _norVarName = "nor"
        for currentVar in _var:
            _norVarName += "_%s"%(currentVar)
           
        if ilpSolver=="Gurobi":
            _norVar = m.addVar(vtype=GRB.BINARY, name=_norVarName)
        elif ilpSolver == "GEKKO":
            _norVar = m.Var(0, lb=0, ub=1, integer=True, name=_norVarName)
    
        for currentVar in _var:
            if ilpSolver=="Gurobi":
                m.addConstr(currentVar, GRB.LESS_EQUAL, 1 - _norVar)
            elif ilpSolver == "GEKKO":
                m.Equation(currentVar + _norVar  <= 1)
        
        if ilpSolver=="Gurobi":
            _varSumLinExpr = LinExpr()
            for currentVar in _var:
                _varSumLinExpr.addTerms(1.0, currentVar)
            
            m.addConstr(_varSumLinExpr, GRB.GREATER_EQUAL, 1 - _norVar)
        elif ilpSolver == "GEKKO":
            m.Equation(m.sum(list(_var)) + _nandVar >= 1)
    
        return _norVar
    
    # Exclusive Disjunction
    def xorVar(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _xorVar = m.addVar(vtype=GRB.BINARY, name="xor_%s_%s"%(_var1, _var2))
            
            m.addConstr(_var1 + _var2 + _xorVar, GRB.LESS_EQUAL, 2)
            m.addConstr(-_var1 - _var2 + _xorVar, GRB.LESS_EQUAL, 0)
            m.addConstr(_var1 - _var2 + _xorVar, GRB.GREATER_EQUAL, 0)
            m.addConstr(-_var1 + _var2 + _xorVar, GRB.GREATER_EQUAL, 0)
        elif ilpSolver == "GEKKO":
            _xorVar = m.Var(0, lb=0, ub=1, integer=True, name="xor_%s_%s"%(_var1, _var2))
            
            m.Equation(_var1 + _var2 + _xorVar <= 2)
            m.Equation(-_var1 - _var2 + _xorVar <= 0)
            m.Equation(_var1 - _var2 + _xorVar >= 0)
            m.Equation(-_var1 + _var2 + _xorVar >= 0)
            
        return _xorVar
    
    # Implication
    def ifVar(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _ifVar = m.addVar(vtype=GRB.BINARY, name="if_%s_%s"%(_var1, _var2))
            
            m.addConstr(1 - _var1 , GRB.LESS_EQUAL, _ifVar)
            m.addConstr(_var2 , GRB.LESS_EQUAL, _ifVar)
            m.addConstr(1 - _var1 + _var2, GRB.GREATER_EQUAL, _ifVar)
        elif ilpSolver == "GEKKO":
            _ifVar = m.Var(0, lb=0, ub=1, integer=True, name="if_%s_%s"%(_var1, _var2))
            
            m.Equation(_var1 + _ifVar >= 1)
            m.Equation(_var2 - _ifVar <= 0)
            m.Equation(_ifVar + _var1 - _var2 <= 1)
            
        return _ifVar
           
    # Equivalence 
    def epqVar(m, _var1, _var2):
        if ilpSolver=="Gurobi":
            _epqVar = m.addVar(vtype=GRB.BINARY, name="epq_%s_%s"%(_var1, _var2))
            
            m.addConstr(_var1 + _var2 - _epqVar, GRB.LESS_EQUAL, 1)
            m.addConstr(_var1 + _var2 + _epqVar , GRB.GREATER_EQUAL, 1)
            m.addConstr(-_var1 + _var2 + _epqVar, GRB.LESS_EQUAL, 1)
            m.addConstr(_var1 - _var2 + _epqVar, GRB.LESS_EQUAL, 1)
        elif ilpSolver == "GEKKO":
            _epqVar = m.Var(0, lb=0, ub=1, integer=True, name="epq_%s_%s"%(_var1, _var2))
            
            m.Equation(_var1 + _var2 - _epqVar <= 1)
            m.Equation(_var1 + _var2 + _epqVar >= 1)
            m.Equation(-_var1 + _var2 + _epqVar <= 1)
            m.Equation(_var1 - _var2 + _epqVar <= 1)
     
        return _epqVar
    
    def main() :
        # Create a new Gurobi model
        m = Model("andVarTest")
        m.Params.InfUnbdInfo = 1
        m.params.outputflag = 1
        
        _var1 = m.addVar(vtype=GRB.BINARY, name="_var1")
        _var2 = m.addVar(vtype=GRB.BINARY, name="_var2")
    
        m.addConstr(_var1, GRB.EQUAL, 1)
        m.addConstr(_var2, GRB.EQUAL, 1)
    
        _andVar = andVar(m, _var1, _var2)
        m.update()
    
        Q = LinExpr()
        Q += _andVar
        m.setObjective(Q, GRB.MAXIMIZE)
    
        m.update()
        
        print(m.getObjective()) 
        for const in m.getConstrs():
            print(m.getConstrByName(const.ConstrName))
            
        m.optimize()
           
        print(m.printStats())
    
        if m.status == GRB.Status.OPTIMAL:
            print('Optimal solution was found')
        elif m.status == GRB.Status.INFEASIBLE:
            print('Model was proven to be infeasible.')
            exit()
        elif m.status == GRB.Status.INF_OR_UNBD:
            print('Model was proven to be infeasible or unbound.')
            exit()
        elif m.status == GRB.Status.UNBOUNDED:
            print('Model was proven to be unbound.')
            exit()
        else:
            print('Optimal solution not was found - error code %i'%(m.status))
            exit()
    
        m.update()
        m.printAttr('x') 
        
        #print(_andVar.getAttr(GRB.Attr.X))
    
    if __name__ == '__main__' :
        main()


from collections import OrderedDict, namedtuple
from itertools import chain

if __package__ is None or __package__ == '':
    from base import BaseGraphTree
    from property import Property
else:
    from .base import BaseGraphTree
    from .property import Property


@BaseGraphTree.localize_namespace
class Graph(BaseGraphTree):
    def __init__(self, name=None, ontology=None, iri=None, local=None, auto_constraint=None, reuse_model=False):
        BaseGraphTree.__init__(self, name)
        if ontology is None:
            self.ontology = (iri, local)
        elif isinstance(ontology, Graph.Ontology):
            self.ontology = ontology
        elif isinstance(ontology, str):
            self.ontology = (ontology, local)
        self.auto_constraint = auto_constraint
        self.reuse_model = reuse_model
        self._concepts = OrderedDict()
        self._logicalConstrains = OrderedDict()
        self._relations = OrderedDict()

    def __iter__(self):
        yield from BaseGraphTree.__iter__(self)
        yield from self._concepts
        yield from self._relations

    @property
    def ontology(self):
        return self._ontology

    @property
    def auto_constraint(self):
        if self._auto_constraint is None and self.sup:
            return self.sup.auto_constraint
        return self._auto_constraint or False  # if None, return False instead

    @auto_constraint.setter
    def auto_constraint(self, value):
        self._auto_constraint = value

    @ontology.setter
    def ontology(self, ontology):
        if isinstance(ontology, Graph.Ontology):
            self._ontology = ontology
        elif isinstance(ontology, str):
            self._ontology = Graph.Ontology(ontology)
        elif isinstance(ontology, tuple) and len(ontology) == 2:
            self._ontology = Graph.Ontology(*ontology)

    def get_properties(self, *tests):
        def func(node):
            if isinstance(node, Property):
                for test in tests:
                    if not test(node):
                        return None
                else:
                    return node
            return None
        return list(self.traversal_apply(func))

    def get_sensors(self, *tests):
        return list(chain(*(prop.find(*tests) for prop in self.get_properties())))

    def get_apply(self, name):
        if name in self.concepts:
            return self.concepts[name]
        if name in self.relations:
            return self.relations[name]
        return BaseGraphTree.get_apply(self, name)

    def set_apply(self, name, sub):
        if __package__ is None or __package__ == '':
            from concept import Concept
            from relation import Relation
        else:
            from .concept import Concept
            from .relation import Relation
        # TODO: what if a concept has same name with a subgraph?
        if isinstance(sub, Graph):
            BaseGraphTree.set_apply(self, name, sub)
        elif isinstance(sub, Concept):
            self.concepts[name] = sub
        elif isinstance(sub, Relation):
            self.relations[name] = sub
        else:
            # TODO: what are other cases
            pass


    def visualize(self, filename, open_image=False):
        import graphviz
        concept_graph = graphviz.Digraph(name=f"{self.name}")
        concept_graph.attr(label=f"Graph: {self.name}") 

        for concept_name, concept in self.concepts.items():
            concept_graph.node(concept_name)

        for subgraph_name, subgraph in self.subgraphs.items():
            sub_graph_viz = subgraph.visualize(filename=None)
            sub_graph_viz.name = 'cluster_' + sub_graph_viz.name
            concept_graph.subgraph(sub_graph_viz)

        for relation_name, relation in self.relations.items():
            if not relation.name.endswith('reversed') and not ('not_a' in relation.name):
                # add case for HasA
                concept_graph.edge(relation.src.name, relation.dst.name, label=relation.__class__.__name__)

        if filename is not None:
            concept_graph.render(filename, format='png', view=open_image)
        else:
            return concept_graph

    @property
    def subgraphs(self):
        return OrderedDict(self)

    @property
    def concepts(self):
        return self._concepts

    @property
    def logicalConstrains(self):
        return self._logicalConstrains

    @property
    def logicalConstrainsRecursive(self):
        def func(node):
            if isinstance(node, Graph):
                yield from node.logicalConstrains.items()
        yield from self.traversal_apply(func)

    @property
    def relations(self):
        return self._relations

    def what(self):
        wht = BaseGraphTree.what(self)
        wht['concepts'] = dict(self.concepts)
        return wht

    # NB: for namedtuple, defaults are right-most first
    # `local` default to None,
    # python 3.7+
    # Ontology = namedtuple('Ontology', ('iri', 'local'), defaults=(None,))
    # python 2.6+
    Ontology = namedtuple('Ontology', ('iri', 'local'))
    Ontology.__new__.__defaults__ = (None,)



import logging

dnConfig = {
    # Logging configuration for DataNode and DataNoteBuilder
    'ifLog': True,
    'log_name' : 'dataNode', 
    'log_level' : logging.INFO,
    'log_filename' : 'logs/datanode.log',
    'log_filesize' : 5*1024*1024*1024,
    'log_backupCount' : 0,
    'log_fileMode' : 'w'
}


from domiknows.sensor.pytorch.relation_sensors import EdgeSensor
import torch
from typing import Any


class SpacyTokenizer(EdgeSensor):
    def __init__(self, *pres, relation, edges=None, label=False, device='auto', spacy=None):
        super().__init__(*pres, relation=relation, edges=edges, label=label, device=device)
        if not spacy:
            raise ValueError('You should select a default Spacy Pipeline')
        self.nlp = spacy

    def forward(self,) -> Any:
        text = self.nlp(self.inputs[0])
        return [token for token in text]



from .graph import Graph
from .concept import Concept, EnumConcept
from .relation import Relation
from .logicalConstrain import LogicalConstrain, V
from .logicalConstrain import andL, nandL, orL, ifL, norL, xorL, epqL, notL, existsL, atLeastL, atMostL, exactL, eqL, fixedL, existsAL, atLeastAL, atMostAL, exactAL
from .property import Property
from .trial import Trial
from .dataNode import DataNode, DataNodeBuilder


import abc
import inspect
from collections import Counter, defaultdict, OrderedDict
from contextlib import contextmanager
from threading import Lock
from itertools import chain
import pprint
from .utils import entuple, singleton, hide_inheritance


class Scoped(object):
    '''
    A lock that count all trials of acquiring and require same number of releasing to really unlock.
    Who else need this...? Well, let me leave it nested inside then...
    '''
    class LevelLock(object):
        # note: threading.Lock is a `builtin_function_or_method`, not a class!
        def __init__(self):
            self.__lock = Lock()
            self.__level = 0

        @property
        def level(self):
            return self.__level

        def acquire(self, blocking=False):  # in this setting, you don't usually want blocking
            retval = self.__lock.acquire(blocking)
            self.__level += 1
            return retval

        def release(self):
            self.__level -= 1
            if self.__level == 0:
                self.__lock.release()

        def __repr__(self):
            return repr(self.__level)

    __locks = defaultdict(LevelLock)

    def __init__(self, blocking=False):
        self._blocking = blocking

    @classmethod
    def clear(cls):
        cls.__context.__locks.clear()

    @contextmanager
    def scope(self, blocking=None):
        if blocking is None:
            blocking = self._blocking
        lock = type(self).__context.__locks[self.scope_key]
        try:
            yield lock.acquire(self._blocking)
        finally:
            lock.release()

    @property
    def scope_key(self):
        return type(self)

    @staticmethod
    def class_scope(cls_):
        # cls_ is not caller (Scoped) class! This is a static function
        cls_.scope_key = Scoped.scope_key
        return cls_

    @staticmethod
    def instance_scope(cls_):
        # cls_ is not caller (Scoped) class! This is a static function
        def scope_key(self_):
            # claim recursive check at instance level so it is possible to show properties of the same type
            return self_
        cls_.scope_key = property(scope_key)
        return cls_


Scoped._Scoped__context = Scoped


class Named(Scoped):
    def __init__(self, name):
        super().__init__(blocking=False)
        self.name = name

    def __repr__(self):
        cls = type(self)
        with self.scope() as detailed:
            if detailed and callable(getattr(self, 'what', None)):
                repr_str = '{class_name}(name=\'{name}\', what={what})'.format(class_name=cls.__name__,
                                                                               name=self.name,
                                                                               what=pprint.pformat(self.what(), width=8))
            else:
                repr_str = '{class_name}(name=\'{name}\')'.format(class_name=cls.__name__,
                                                                  name=self.name)

        return repr_str

    def __str__(self):
        return str(self.name)


class AutoNamed(Named):
    _names = Counter()
    _objs = dict()

    @classmethod
    def clear(cls):
        super().clear()
        cls._names.clear()
        cls._objs.clear()

    @staticmethod
    def localize_namespace(cls_):
        # cls_ is not caller (AutoNamed) class! This is a static function
        cls_._names = Counter()
        cls_._objs = dict()
        return cls_

    @classmethod
    def get(cls, name, value=None):
        return cls._objs.get(name, value)

    @classmethod
    def suggest_name(cls):
        return cls.__name__.lower()

    def assign_suggest_name(self, name=None):
        cls = type(self)
        if name is None:
            name = cls.suggest_name()
        if cls._names[name] > 0:
            while True:
                name_attempt = '{}-{}'.format(name, cls._names[name])
                cls._names[name] += 1
                if cls._names[name_attempt] == 0:
                    name = name_attempt
                    break  # while True
        assert cls._names[name] == 0
        cls._names[name] += 1
        self.name = name
        cls._objs[name] = self


    def __init__(self, name=None):
        Named.__init__(self, name)  # temporary name may apply
        self.assign_suggest_name(name)

    @staticmethod
    def named_singleton(cls_):
        # cls_ is not caller (AutoNamed) class! This is a static function
        if not issubclass(cls_, AutoNamed):
            raise TypeError('named_singleton can be applied to subclasses of AutoNamed,' +
                            ' but MRO of {} is given.'.format(cls_.mro()))

        def getter(name=None):
            return cls_.get(name, None)

        def setter(obj):
            pass  # AutoNamed save them already

        cls_ = singleton(cls_, getter, setter)
        return cls_


class NamedTreeNode(Named):
    _context = []

    @staticmethod
    def localize_context(cls_):
        # cls_ is not caller (NamedTreeNode) class! This is a static function
        cls_._context = []
        return cls_

    @classmethod
    def share_context(cls, cls_):
        # cls is the caller (NamedTreeNode) class
        # cls_ is another class that need to share the same context
        cls_._context = cls._context
        return cls_

    @classmethod
    def clear(cls):
        super().clear()
        cls._context.clear()

    @classmethod
    def default(cls):
        if cls._context:
            return cls._context[-1]
        return None

    def __init__(self, name=None):
        super().__init__(name)
        cls = type(self)
        self._sup = None
        self.attach_to_context()

    def attach_to_context(self, name=None):
        cls = type(self)
        if len(cls._context) == 0:
            context = None
        else:
            context = cls._context[-1]
            context.attach(self, name)

    def attached(self, sup):
        pass

    def __enter__(self):
        cls = type(self)
        #self.sup = cls._context
        self.attach_to_context()  # TODO: this could lead to switching sup to context
        cls._context.append(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        cls = type(self)
        last = cls._context.pop()
        assert last is self

    @property
    def sup(self):
        return self._sup

    @property
    def sups(self):
        if self.sup is not None:
            yield self.sup
            yield from self.sup.sups

    @property
    def fullname(self):
        return self.get_fullname()

    def get_fullname(self, delim='/'):
        if self.sup is None:
            return self.name
        return self.sup.get_fullname(delim) + delim + self.name

    def what(self):
        return {'sup': self.sup}


@Scoped.instance_scope
class NamedTree(NamedTreeNode, OrderedDict):
    def __repr__(self):
        with hide_inheritance(NamedTree, dict):
            # prevent pprint over optimize OrderedDict(dict) on NamedTree instance
            return super().__repr__()

    def __hash__(self):  # NB: OrderedDict is unhashable.
        # dict override __hash__ by setting it to None. Using super().__hash__() gives problem.
        return NamedTreeNode.__hash__(self)

    def __eq__(self, other): # NB: OrderedDict has __eq__, empty == empty, which is not what we expected
        # dict override __eq__ in an waired way also
        return NamedTreeNode.__eq__(self, other)

    def __init__(self, name=None):
        super().__init__(name)
        super(NamedTreeNode, self).__init__(self.name)

    def attach(self, sub, name=None):
        # resolve and prevent recursive definition
        if sub is self or sub in self.sups:
            raise ValueError('Recursive definition detected for attaching {} to {} with sups {}'.format(
                sub.name, self.name, [sup.name for sup in self.sups]))
        if isinstance(sub, NamedTreeNode):
            if sub.sup is not None:
                sub.sup.detach(sub)
            sub._sup = self # FIXME: using private method
        if isinstance(sub, Named):
            if name is None:
                name = sub.name
            self[name] = sub
        else:
            raise TypeError(
                'Attach Named instance to NamedTree, {} instance given.'.format(type(sub)))
        sub.attached(self)

    def detach(self, sub=None, all=False):
        if sub is None:
            if all:
                for sub in dict(self).values():
                    # dict() makes a shallow copy to avoid runtime error of change
                    self.detach(sub)
                #return self.clear()
            else:
                # detach all leaves
                for sub in dict(self).values():
                    # dict() makes a shallow copy to avoid runtime error of change
                    if isinstance(sub, NamedTree):
                        sub.detach()
                    else:
                        self.detach(sub)
            return
        # detach specific sub
        # NB: sub.name is not reliable!
        for key, value in dict(self).items():
            # dict() makes a shallow copy to avoid runtime error of change
            if value == sub:
                del self[key]
                sub._sup = None # TODO: what else to have?

    def traversal_apply(self, func, filter_fn=lambda x: x is not None, order='pre', first='depth'):
        if order.lower() not in ['pre', 'post']:
            raise ValueError('Options for order are pre or post, {} given.'.format(order))
        if first.lower() not in ['depth', 'breadth']:
            raise ValueError('Options for first are depth or breadth, {} given.'.format(first))

        to_traversal = [self, ]
        to_apply = []

        while to_traversal:
            if first.lower() == 'depth':
                current = to_traversal.pop()
            else:
                current = to_traversal.pop(0)
            to_apply.append(current)

            if order.lower() == 'pre':
                current_apply = to_apply.pop()
                if inspect.isgeneratorfunction(func):
                    # func: yield
                    if filter_fn:
                        yield from filter(filter_fn, func(current_apply))
                    else:
                        yield from func(current_apply)
                else:
                    # func: return
                    retval = func(current_apply)
                    if not filter_fn or filter_fn(retval):
                        yield retval

            if isinstance(current, NamedTree):
                subs = [current[name] for name in current] # compatible to subclasses that override the iterator
                if (first.lower() == 'depth' and order.lower() == 'pre') or (
                    first.lower() == 'breadth' and order.lower() == 'post'):
                    subs = reversed(subs)

                to_traversal.extend(subs)

        if order.lower() == 'post':
            while to_apply:
                current_apply = to_apply.pop()
                if inspect.isgeneratorfunction(func):
                    # func: yield
                    if filter_fn:
                        yield from filter(filter_fn, func(current_apply))
                    else:
                        yield from func(current_apply)
                else:
                    # func: return
                    retval = func(current_apply)
                    if not filter_fn or filter_fn(retval):
                        yield retval

    def extract_name(self, *names, delim='/', trim=True):
        if isinstance(names[0], str):
            name0s = names[0].split(delim)
            name = name0s[0]
            if trim:
                name = name.strip()
            names = list(chain(name0s[1:], names[1:]))
            # match typed name string
            if name[0] == '<' and name[-1] == '>':
                for key in self:
                    if key.name == name[1:-1]:
                        name = key
                        break
        else:
            name = names[0]
            names = names[1:]
        return name, names

    def cutGraphName(self, names):
        gNames = self._names.copy()
        index = 0
        
        for i, name in enumerate(names):
            index = i
            if (name in gNames) and (gNames[name] > 0):
                gNames[name] -= 1
            else:
                break
            
        return index# names[index:]
        
    def parse_query_apply(self, func, *names, delim='/', trim=True):
        name, names = self.extract_name(*names, delim=delim, trim=trim)
        if names:
            return self[name].parse_query_apply(func, *names, delim=delim, trim=trim)
        return func(self, name)

    def get_apply(self, name):
        return OrderedDict.__getitem__(self, name)

    def get_sub(self, *names, delim='/', trim=True):
        return self.parse_query_apply(lambda s, name: s.get_apply(name), *names, delim=delim, trim=trim)

    def set_apply(self, name, sub):
        OrderedDict.__setitem__(self, name, sub)

    def set_sub(self, *names, sub, delim='/', trim=True):
        # NB: sub is keyword arg because it is after a list arg
        return self.parse_query_apply(lambda s, k: s.set_apply(k, sub), *names, delim=delim, trim=trim)

    def del_apply(self, name):
        return OrderedDict.__delitem__(self, name)

    def del_sub(self, *names, delim='/', trim=True):
        return self.parse_query_apply(lambda s, name: s.del_apply(name), *names, delim=delim, trim=trim)

    def __getitem__(self, name):
        try:
            return self.get_sub(*entuple(name))
        except KeyError as e:
            raise type(e)(name)

    def __setitem__(self, name, obj):
        return self.set_sub(*entuple(name), sub=obj)

    def __delitem__(self, name):
        try:
            return self.del_sub(*entuple(name))
        except KeyError as e:
            raise type(e)(name)

    def what(self):
        wht = NamedTreeNode.what(self)
        wht['subs'] = dict(self)
        return wht



if __package__ is None or __package__ == '': 
    from domiknows.solver.ilpConfig import ilpConfig
    from domiknows.solver.ilpOntSolver import ilpOntSolver
else:
    from .ilpConfig import ilpConfig 
    from .ilpOntSolver import ilpOntSolver

class dummyILPOntSolver(ilpOntSolver):
    def __init__(self, graph, ontologiesTuple, _ilpConfig=ilpConfig) -> None:
        super().__init__(graph, ontologiesTuple, _ilpConfig=ilpConfig)
        self.myIlpBooleanProcessor = None
        
    def calculateILPSelection(self, phrase, fun=None, epsilon = 0.00001, graphResultsForPhraseToken=None, graphResultsForPhraseRelation=None, graphResultsForPhraseTripleRelation=None):
        self.myLogger.info('Returning unchanged results')
        return graphResultsForPhraseToken, graphResultsForPhraseRelation, graphResultsForPhraseTripleRelation


import logging

config = {
    # Logging configuration for regrTimer
    'ifLog': True,
    'log_name' : 'regrTimer', 
    'log_level' : logging.INFO,
    'log_filename' : 'logs/regrTimer',
    'log_filesize' : 5*1024*1024*1024,
    'log_backupCount' : 5,
    'log_fileMode' : 'a'
}


import logging
import warnings
from collections import OrderedDict

import numpy as np
import torch

from ...graph import DataNodeBuilder, DataNode
from ..metric import MetricTracker, MacroAverageTracker

class LossModel(torch.nn.Module):
    logger = logging.getLogger(__name__)

    def __init__(self, graph, 
                 tnorm=DataNode.tnormsDefault, 
                 sample = False, sampleSize = 0, sampleGlobalLoss = False, device='auto'):
        super().__init__()
        self.graph = graph
        self.build = True
        
        self.tnorm = tnorm
        self.device = device
        
        self.sample = sample
        self.sampleSize = sampleSize
        self.sampleGlobalLoss = sampleGlobalLoss
        
        self.constr = OrderedDict(graph.logicalConstrainsRecursive)
        nconstr = len(self.constr)
        if nconstr == 0:
            warnings.warn('No logical constraint detected in the graph. '
                          'PrimalDualModel will not generate any constraint loss.')
            
        self.lmbd = torch.nn.Parameter(torch.empty(nconstr))
        self.lmbd_p = torch.empty(nconstr)  # none parameter
        self.lmbd_index = {}
        
        for i, (key, lc) in enumerate(self.constr.items()):
            self.lmbd_index[key] = i
            p = float(lc.p) / 100.
            if p == 1:
                p = 0.999999999999999
            self.lmbd_p[i] = -np.log(1 - p)  # range: [0, inf)
            
        self.reset_parameters()
        self.loss = MacroAverageTracker(lambda x:x)

    def reset_parameters(self):
        torch.nn.init.constant_(self.lmbd, 1.)

    def to(self, device):
        super().to(device=device)
        if self.device is not None:
            if hasattr(self, 'lmbd_p'):
                self.lmbd_p = self.lmbd_p.to(device)

    def reset(self):
        if isinstance(self.loss, MetricTracker):
            self.loss.reset()

    def get_lmbd(self, key):
        return self.lmbd[self.lmbd_index[key]].clamp(max=self.lmbd_p[self.lmbd_index[key]])

    def forward(self, builder, build=None):
        if build is None:
            build = self.build
            
        if not build and not isinstance(builder, DataNodeBuilder):
            raise ValueError('PrimalDualModel must be invoked with `build` on or with provided DataNode Builder.')
        
        if (builder.needsBatchRootDN()):
            builder.addBatchRootDN()
        datanode = builder.getDataNode(device=self.device)
        
        # Call the loss calculation returns a dictionary, keys are matching the constraints
        constr_loss = datanode.calculateLcLoss(tnorm=self.tnorm, sample=self.sample, sampleSize = self.sampleSize)

        lmbd_loss = []
        if self.sampleGlobalLoss and constr_loss['globalLoss']:
            globalLoss = constr_loss['globalLoss']
            self.loss['globalLoss'](globalLoss)
            lmbd_loss = torch.tensor(globalLoss, requires_grad=True)
        else:
            for key, loss in constr_loss.items():
                if key not in self.constr:
                    continue
                
                if loss['lossTensor'] != None:
                    loss_value = loss['lossTensor'].clamp(min=0)
                    loss_nansum = loss_value[loss_value==loss_value].sum()
                    loss_ = self.get_lmbd(key) * loss_nansum
                    self.loss[key](loss_)
                    lmbd_loss.append(loss_)    
               
            lmbd_loss = sum(lmbd_loss)
        
        # (*out, datanode, builder)
        return lmbd_loss, datanode, builder
    
class PrimalDualModel(LossModel):
    logger = logging.getLogger(__name__)

    def __init__(self, graph, tnorm=DataNode.tnormsDefault, device='auto'):
        super().__init__(graph, tnorm=tnorm, device=device)
        
class SampleLosslModel(torch.nn.Module):
    logger = logging.getLogger(__name__)

    # def __init__(self, graph, sample = False, sampleSize = 0, sampleGlobalLoss = False):
    #     super().__init__(graph, sample=sample, sampleSize=sampleSize, sampleGlobalLoss=sampleGlobalLoss)

    def __init__(self, graph, 
                 tnorm=DataNode.tnormsDefault, 
                 sample = False, sampleSize = 0, sampleGlobalLoss = False, device='auto'):
        super().__init__()
        self.graph = graph
        self.build = True
        
        self.tnorm = tnorm
        self.device = device
        
        self.sample = sample
        self.sampleSize = sampleSize
        self.sampleGlobalLoss = sampleGlobalLoss
        
        self.constr = OrderedDict(graph.logicalConstrainsRecursive)
        nconstr = len(self.constr)
        if nconstr == 0:
            warnings.warn('No logical constraint detected in the graph. '
                          'PrimalDualModel will not generate any constraint loss.')
            
        self.lmbd = torch.nn.Parameter(torch.zeros(nconstr).float())
        self.lmbd_index = {}

        self.iter_step = 0
        self.warmpup = 80
        
        for i, (key, lc) in enumerate(self.constr.items()):
            self.lmbd_index[key] = i
            
        self.reset_parameters()
        self.loss = MacroAverageTracker(lambda x:x)

    def reset_parameters(self):
        torch.nn.init.constant_(self.lmbd, 0.0)

    def reset(self):
        if isinstance(self.loss, MetricTracker):
            self.loss.reset()

    def get_lmbd(self, key):
        if self.lmbd[self.lmbd_index[key]] < 0:
            with torch.no_grad():
                self.lmbd[self.lmbd_index[key]] = 0
        return self.lmbd[self.lmbd_index[key]]

    def forward(self, builder, build=None):
        if build is None:
            build = self.build
        self.iter_step += 1
            
        if not build and not isinstance(builder, DataNodeBuilder):
            raise ValueError('PrimalDualModel must be invoked with `build` on or with provided DataNode Builder.')
        
        if (builder.needsBatchRootDN()):
            builder.addBatchRootDN()
        
#         self.loss.reset()

        datanode = builder.getDataNode(device=self.device)
        
        # Call the loss calculation returns a dictionary, keys are matching the constraints
        constr_loss = datanode.calculateLcLoss(tnorm=self.tnorm, sample=self.sample, sampleSize = self.sampleSize, sampleGlobalLoss = self.sampleGlobalLoss)
        import math
        lmbd_loss = []
        replace_mul = False
        
        key_losses = dict()
        for key, loss in constr_loss.items():
            if key not in self.constr:
                continue
            # loss_value = loss['loss']
            epsilon = 0.0
            key_loss = 0
            new_eps = 0.01
            for i, lossTensor in enumerate(loss['lossTensor']):
                lcSuccesses = loss['lcSuccesses'][i]
                if self.sampleSize == -1:
                    sample_info = [val_ for key, val in loss['sampleInfo'].items() for val_ in val if len(val_)]
                    sample_info = [val[i][1] for val in sample_info]
                    sample_info = torch.stack(sample_info).t()
                    unique_output, unique_inverse, counts = torch.unique(sample_info, return_inverse=True, dim=0, return_counts=True)
                    _, ind_sorted = torch.sort(unique_inverse, stable=True)
                    cum_sum = counts.cumsum(0)
                    cum_sum = torch.cat((torch.tensor([0]).to(counts.device), cum_sum[:-1]))
                    first_indicies = ind_sorted[cum_sum]
                    assert lcSuccesses.sum().item() != 0
                    tidx = (lcSuccesses == 1).nonzero().squeeze(-1)
                    unique_selected_indexes = torch.tensor(np.intersect1d(first_indicies.cpu().numpy(), tidx.cpu().numpy()))
                    if unique_selected_indexes.shape:
                        loss_value = lossTensor[unique_selected_indexes].sum()
                        loss_ = -1 * torch.log(loss_value)
                        key_loss += loss_

                    else:
                        loss_ = 0
                    
                else:
                    if constr_loss["globalSuccessCountet"] > 0:
                        lcSuccesses = constr_loss["globalSuccesses"]
                    if lossTensor.sum().item() != 0:
                        tidx = (lcSuccesses == 1).nonzero().squeeze(-1)
                        true_val = lossTensor[tidx]
                        
                        if true_val.sum().item() != 0: 
                            if not replace_mul:
                                loss_value = true_val.sum() / lossTensor.sum()
                                loss_value = epsilon - ( -1 * torch.log(loss_value) )
                                # loss_value = -1 * torch.log(loss_value) 
                                if self.iter_step < self.warmpup:
                                    with torch.no_grad():
                                        min_val = loss_value
                                else:
                                    min_val = -1
                                # min_val = -1
                                # with torch.no_grad():
                                #     min_val = loss_value
                                loss_ = min_val * loss_value
                                key_loss += loss_
                            else:
                                loss_value = true_val.logsumexp(dim=0) - lossTensor.logsumexp(dim=0)
                                key_loss += -1 * loss_value

                        else:
                            loss_ = 0
                        
                        

                        # if loss['lossTensor'].nansum().item() <= 1:
                        #     loss_value = loss['lossTensor']
                        #     # loss_value = loss_value[loss_value.nonzero()].squeeze(-1)
                        # else:
                        #     _idx = []
                        #     for i in torch.unique(loss['lossTensor']):
                        #         _idx.append((loss['lossTensor'] == i.item()).nonzero(as_tuple=True)[0][0].item())
                        #     loss_value = loss['lossTensor'][_idx]
                        #     # loss_value = torch.unique(loss['lossTensor'])

                        # loss_value = torch.log(loss_value.sum())
                        # loss_ = -1 * (loss_value)
                        # self.loss[key](loss_)
                        # lmbd_loss.append(loss_) 
                        
                    else:
                        loss_ = 0

            epsilon = 1e-2
            if self.sampleSize != -1:
                key_loss = max(key_loss - epsilon, 0) 
            if key_loss != 0:  
                key_losses[key] = key_loss
                # self.loss[key](key_loss)
                # lmbd_loss.append(key_loss) 
                    
        all_losses = [key_losses[key] for key in key_losses]
        if all_losses:
            all_losses = torch.stack(all_losses)

            satisfied_num = len( set(constr_loss.keys()) - set(key_losses.keys()) )
            unsatisfied_num = len(set(constr_loss.keys())) - satisfied_num
            #self.logger.info(f'-- number of satisfied constraints are {satisfied_num}')
            #self.logger.info(f'-- number of unstatisfied constraints are {unsatisfied_num}')
            for key in key_losses:
                if self.sampleSize != -1:
                    if replace_mul:
                        loss_val = (key_losses[key] / all_losses.sum()) * key_losses[key]
                    else:
                        loss_val = key_losses[key]
                else:
                    loss_val = key_losses[key]

                self.loss[key](loss_val)
                lmbd_loss.append(loss_val) 
                
            lmbd_loss = sum(lmbd_loss)
        
        # (*out, datanode, builder)
        # self.logger.info(f'-- lmbd_loss is {lmbd_loss}')
        return lmbd_loss, datanode, builder



import logging
from itertools import product, permutations

from domiknows.utils import isbad

from ..session.solver_session import SolverSession


class Constructor():
    logger = logging.getLogger(__name__)

    def __init__(self, lazy_not=True, self_relation=True):
        self.lazy_not = lazy_not
        self.self_relation = self_relation

    def get_predication(self, predicate, idx, negative=False):
        raise NotImplementedError

    def isskip(self, value):
        return isbad(value)

    def candidates(self, data, *predicates_list):
        candidates = {} # concept -> [(object,...), ...]
        if self.self_relation:
            gen = lambda enum_data, arity: product(enum_data, repeat=arity)
        else:
            gen = lambda enum_data, arity: permutations(enum_data, r=arity)
        for arity, predicates in enumerate(predicates_list, 1):
            for concept in predicates:
                #assert concept not in candidates
                # last one change first (c-order)
                # abc rep=3 -> aaa, aab, aac, aba, abb, abc, ...
                candidates[concept] = tuple(gen(enumerate(data), arity))
        return candidates

    def variables(self, session, candidates, *predicates_list):
        variables = {} # (concept, (object,...)) -> variable
        predictions = {} # (concept, (object,...)) -> prediction
        variables_not = {} # (concept, (x,...)) -> variable
        predictions_not = {} # (concept, (x,...)) -> prediction

        # add variables
        self.logger.debug('add variables')
        for predicates in predicates_list:
            for concept, predicate in predicates.items():
                self.logger.debug('for %s', concept.name)
                self.logger.debug(predicate)
                for x in candidates[concept]: # flat: C-order -> last dim first!
                    idx, _ = zip(*x)
                    prediction = self.get_predication(predicate, idx)
                    if self.isskip(prediction): continue
                    var = session.var(
                        session.VTYPE.BIN, 0, 1,
                        name='{}_{}'.format(concept.name, str(x)))
                    self.logger.debug(' - add %s', var)
                    variables[concept, x] = var
                    predictions[concept, x] = prediction

        if self.lazy_not:
            self.logger.debug('lazy negative')

            # add variables
            self.logger.debug('lazy negative add variables')
            for predicates in predicates_list:
                for concept, predicate in predicates.items():
                    self.logger.debug('for %s', concept.name)
                    for x in candidates[concept]:
                        idx, _ = zip(*x)
                        prediction_not = self.get_predication(predicate, idx, negative=True)
                        if self.isskip(prediction_not): continue
                        var = session.var(
                            session.VTYPE.BIN, 0, 1,
                            name='lazy_not_{}_{}'.format(concept.name, str(x)))
                        self.logger.debug(' - add %s', var)
                        variables_not[concept, x] = var
                        predictions_not[concept, x] = prediction_not

        return variables, predictions, variables_not, predictions_not

    def constraints(self, session, candidates, variables, variables_not, *predicates_list):
        constraints = {} # (rel, (object,...)) -> constr
        constraints_not = {} # (rel, (x,...)) -> constr

        # add constraints
        self.logger.debug('add constraints')
        for predicates in predicates_list:
            for concept in predicates:
                self.logger.debug('for %s', concept.name)
                self.logger.debug(' - is_a')
                for rel in concept.is_a():
                    self.logger.debug(' - - %s', rel.name)
                    # A is_a B : A(x) <= B(x)
                    for x in candidates[rel.src]:
                        if (rel.src, x) not in variables: continue
                        if (rel.dst, x) not in variables: continue
                        constr = session.constr(
                            variables[rel.src, x], SolverSession.CTYPE.LE, variables[rel.dst, x],
                            name='{}_{}'.format(rel.name, str(x)))
                        self.logger.debug(' - - add %s', constr)
                        assert (rel, x) not in constraints
                        constraints[rel, x] = constr
                self.logger.debug(' - not_a')
                for rel in concept.not_a():
                    self.logger.debug(' - - %s', rel.name)
                    # A not_a B : A(x) + B(x) <= 1
                    for x in candidates[rel.src]:
                        if (rel.src, x) not in variables: continue
                        if (rel.dst, x) not in variables: continue
                        constr = session.constr(
                            variables[rel.src, x] + variables[rel.dst, x], SolverSession.CTYPE.LE, 1,
                            name='{}_{}'.format(rel.name, str(x)))
                        self.logger.debug(' - - add %s', constr)
                        assert (rel, x) not in constraints
                        constraints[rel, x] = constr
                self.logger.debug(' - has_a')
                for arg_id, rel in enumerate(concept.has_a()): # TODO: need to include indirect ones like sp_tr is a tr while tr has a lm
                    self.logger.debug(' - - %s', rel.name)
                    # A has_a B : A(x,y,...) <= B(x)
                    for xy in candidates[rel.src]:
                        x = xy[arg_id]
                        if (rel.src, xy) not in variables: continue
                        if (rel.dst, (x,)) not in variables: continue
                        constr = session.constr(
                            variables[rel.src, xy], SolverSession.CTYPE.LE, variables[rel.dst, (x,)],
                            name='{}_{}_{}'.format(rel.name, str(xy), str(x)))
                        self.logger.debug(' - - add %s', constr)
                        assert (rel, xy, (x,)) not in constraints
                        constraints[rel, xy, (x,)] = constr

        if self.lazy_not:
            self.logger.debug('lazy negative add constraints')
            for predicates in predicates_list:
                for concept in predicates:
                    self.logger.debug('for %s', concept.name)
                    for x in candidates[concept]:
                        if (concept, x) not in variables: continue
                        if (concept, x) not in variables_not: continue
                        constr = session.constr(
                            variables[concept, x] + variables_not[concept, x], SolverSession.CTYPE.EQ, 1,
                            name='lazy_not_{}_{}'.format(concept.name, str(x)))
                        self.logger.debug(' - add %s', constr)
                        constraints_not[concept, x] = constr
    
        return constraints, constraints_not

    def objective(self, candidates, variables, predictions, variables_not, predictions_not, *predicates_list):
        self.logger.debug('set objective')
        objective = None
        for predicates in predicates_list:
            for concept in predicates:
                for x in candidates[concept]:
                    if (concept, x) not in variables: continue
                    objective += variables[concept, x] * predictions[concept, x]
        if self.lazy_not:
            for predicates in predicates_list:
                for concept in predicates:
                    for x in candidates[concept]:
                        if (concept, x) not in variables_not: continue
                        objective += variables_not[concept, x] * predictions_not[concept, x]
        return objective


class ScoreConstructor(Constructor):
    def get_predication(self, predicate, idx, negative=False):
        if negative:
            return predicate[(*idx, 0)]
        return predicate[(*idx, 1)]


class ProbConstructor(Constructor):
    def __init__(self, lazy_not=False, self_relation=True):
        super().__init__(lazy_not=lazy_not, self_relation=self_relation)

    def get_predication(self, predicate, idx, negative=False):
        if negative:
            return 1 - predicate[idx]
        return predicate[idx]


class BatchMaskProbConstructor(Constructor):
    def get_predication(self, predicate, idx, negative=False):
        value, mask = predicate
        if negative:
            value = 1 - value
        return value[(slice(value.shape[0]),*idx)], mask[(slice(mask.shape[0]),*idx)]

    def isskip(self, value):
        return False


import os
from typing import Dict, List, Tuple, Iterable
from collections import OrderedDict
from torch import Tensor, cuda
from torch.nn import Module
import torch
from allennlp.models import Model
from allennlp.data import Vocabulary
from allennlp.nn.util import get_text_field_mask

from allennlp.training.metrics import CategoricalAccuracy, F1Measure

from .metrics import Epoch, DatasetType

from .. import Graph
from ..trial import Trial
from ..property import Property
from ...utils import prod, get_prop_result
from ...sensor.allennlp import AllenNlpLearner
from ...sensor.allennlp.base import ModuleSensor
from ...sensor.allennlp.sensor import CandidateSensor
from .utils import sequence_cross_entropy_with_logits, structured_perceptron_exact_with_logits


DEBUG_TRAINING = 'REGR_DEBUG' in os.environ and os.environ['REGR_DEBUG']


DataInstance = Dict[str, Tensor]


def update_metrics(
    graph: Graph,
    data: DataInstance,
    metrics: List[Tuple[str, Property, callable]]
) -> DataInstance:
    label_masks = {}
    for sensor in graph.get_sensors(CandidateSensor):
        sensor(data)
        mask = data[sensor.fullname].clone().detach()
        label_masks[mask.shape] = mask
    for metric_name, class_index, prop, metric in metrics:
        label, pred, mask = get_prop_result(prop, data)
        label_mask = label_masks.get(mask.shape)
        if label_mask is not None:
            mask = mask * label_mask.float()
        metric(pred, label, mask)
    return data


class GraphModel(Model):
    def __init__(
        self,
        graph: Graph,
        vocab: Vocabulary,
        balance_factor: float = 0.5,
        label_smoothing: float = 0.1,
        focal_gamma: float = 2.,
        inference_interval: int = 10,
        inference_training_set: bool = False,
        inference_testing_set: bool = True,
        inference_loss: bool = False,
        soft_penalty: float =  1.
    ) -> None:
        super().__init__(vocab)
        self.inference_interval = inference_interval
        self.inference_training_set = inference_training_set
        self.inference_testing_set = inference_testing_set
        self.inference_loss = inference_loss
        self.meta = []
        self.metrics = []
        self.metrics_inferenced = []

        self.balance_factor = balance_factor
        self.label_smoothing = label_smoothing
        self.focal_gamma = focal_gamma
        self.soft_penalty = soft_penalty

        self.graph = graph

        self.meta.append(('epoch', Epoch()))
        self.meta.append(('type', DatasetType()))

        whole_metrics = {
            'Accuracy': CategoricalAccuracy,
        }
        class_metrics = {
            ('P', 'R', 'F1'): F1Measure,
        }

        for prop in self.graph.poi:
            for metric_name, MetricClass in whole_metrics.items():
                self.metrics.append((metric_name, None, prop, MetricClass()))
                self.metrics_inferenced.append((metric_name, None, prop, MetricClass()))
            for metric_name, MetricClass in class_metrics.items():
                for sensor in prop.find(AllenNlpLearner):
                    class_num = prod(sensor.output_dim)
                    if class_num == 2:
                        self.metrics.append((metric_name, None, prop, MetricClass(1)))
                        self.metrics_inferenced.append((metric_name, None, prop, MetricClass(1)))
                    else:
                        for class_index in range(class_num):
                            self.metrics.append((metric_name, class_index, prop, MetricClass(class_index)))
                            self.metrics_inferenced.append((metric_name, class_index, prop, MetricClass(class_index)))

        #i = 0  # TODO: this looks too bad
        for prop in self.graph.poi:
            for sensor in prop.find(ModuleSensor, lambda s: s.module is not None):
                self.add_module(sensor.name, sensor.module)
                #i += 1

    def _need_inference(
        self,
        data: DataInstance
    ) -> bool:
        return self._report_inference(data) or self.inference_loss

    def _report_inference(
        self,
        data: DataInstance
    ) -> bool:
        #import pdb; pdb.set_trace()
        if DEBUG_TRAINING:
            return True
        dataset_type_key = 'dataset_type' # specify in domiknows.graph.allennlp.base.AllenNlpGraph
        if (not self.inference_training_set and
            dataset_type_key in data and
            all(dataset_type == 'train' for dataset_type in data[dataset_type_key])):
            return False
        if (self.inference_testing_set and
            dataset_type_key in data and
            all(dataset_type == 'test' for dataset_type in data[dataset_type_key])):
            return True
        epoch_key = 'epoch_num'  # TODO: this key... is from Allennlp doc
        if epoch_key not in data:
            return True  # no epoch record, then always inference
        epoch = min(data[epoch_key])
        need = ((epoch + 1) % self.inference_interval) == 0  # inference every 10 epoch
        return need

    def _update_metrics(
        self,
        trivial_trial, trial
    ) -> DataInstance:
        for metric_name, metric in self.meta:
            metric(trivial_trial)
        update_metrics(self.graph, trivial_trial, self.metrics)
        if self._need_inference(trivial_trial) and self._report_inference(trivial_trial):
            update_metrics(self.graph, trial, self.metrics_inferenced)
        return trial

    def get_metrics(self, reset: bool=False) -> Dict[str, float]:
        metrics = OrderedDict()

        def add(metric_name, class_index, suffix, prop, metric):
            try:
                out = metric.get_metric(reset)
            except RuntimeError:
                # in case inferenced ones are not called
                # RuntimeError("You never call this metric before.")
                # then pass
                return

            if isinstance(out, Iterable):
                for i, (metric_name_item, out_item) in enumerate(sorted(zip(metric_name, out))):
                    key = prop.sup.name
                    if class_index is not None:
                        class_name = self.vocab.get_token_from_index(class_index, namespace='labels')
                        key += '({})'.format(class_name)
                    if suffix is not None:
                        key += '{}'.format(suffix)
                    key += '-{}'.format(metric_name_item)
                    if i == 0:
                        key = '\n' + key
                    metrics[key] = out_item
            else:
                key = prop.sup.name
                if class_index is not None:
                    class_name = self.vocab.get_token_from_index(class_index, namespace='labels')
                    key += '({})'.format(class_name)
                if suffix is not None:
                    key += '{}'.format(suffix)
                key += '-{}'.format(metric_name)
                key = '\n' + key
                metrics[key] = out

        for metric_name, class_index, prop, metric in self.metrics:
            add(metric_name, class_index, None, prop, metric)
        for metric_name, class_index, prop, metric in self.metrics_inferenced:
            add(metric_name, class_index, '_i', prop, metric)

        max_len = max(len(metric_name.lstrip()) for metric_name in metrics)
        pretty_metrics = OrderedDict()
        for metric_name, metric in sorted(metrics.items(), key=lambda x: x[0].strip()):
            lspace_len = len(metric_name) - len(metric_name.lstrip())
            lspace = metric_name[:lspace_len]
            name = metric_name[lspace_len:]
            pretty_name = '{}{:>{max_len}}'.format(lspace, name, max_len=max_len)
            pretty_metrics[pretty_name] = metric

        for i, (metric_name, metric) in enumerate(self.meta):
            pretty_metrics['{}[ {} ]'.format('\n' if not i else '', metric_name)] = metric.get_metric(reset)
        return pretty_metrics

    def _update_loss(self, trivial_trial, trial):
        trial['loss'] = self.loss_func(trivial_trial, trial)
        return trial

    i = 0
    def forward(
        self,
        **data: DataInstance
    ) -> DataInstance:
        Trial.clear() # reset at every epoch

#         if (self.i % 100 == 0):
#             import gc
#             gc.collect()
#             if cuda.is_available():
#                 cuda.empty_cache()
#             self.i = 1
#         else:
#             self.i += 1

        # make sure every node needed are calculated
        for prop in self.graph.poi:
            for name, sensor in prop.items():
                sensor(data)
        trivial_trial = Trial(data)

        # inference
        with trivial_trial:
            if self._need_inference(trivial_trial):
                trial = Trial()
                self._inference(trial)
            else:
                trial = trivial_trial
        self._update_loss(trivial_trial, trial)
        self._update_metrics(trivial_trial, trial)

        return dict(trial)

    def _inference(
        self,
        data: DataInstance
    ) -> DataInstance:
        #import pdb; pdb.set_trace()
        # process candidates
        label_not_masks = {}
        for sensor in self.graph.get_sensors(CandidateSensor):
            sensor(data)
            mask = data[sensor.fullname].clone().detach()
            label_not_masks[mask.shape] = (1-mask).type(torch.uint8)
        for prop in self.graph.poi:
            # label (b, l)
            # pred  (b, l, c)
            # mask  (b, l)
            label, pred, mask = get_prop_result(prop, data)
            if mask.shape in label_not_masks:
                # label_mask  (b, l)
                label_not_mask = label_not_masks[mask.shape]
                shape = pred.shape
                # (b*l, c)
                pred = pred.clone().detach().view(-1, shape[-1])
                # (b*l, )
                label_not_mask = label_not_mask.view(-1)
                pred[label_not_mask, :] = torch.tensor(float("nan"))
                # (b, l, c)
                pred = pred.view(shape)
                label_not_mask = label_not_mask.view(mask.shape)
                data[prop.fullname] = pred
                for learner in prop.find(AllenNlpLearner):
                    data[learner.fullname] = pred
        #data = inference(self.graph, self.graph.solver, data, self.vocab)
        data = self.graph.solver.inferSelection(self.graph, data)

        return data

    def loss_func(
        self,
        data, data_i
    ) -> DataInstance:
        #import pdb; pdb.set_trace()
        loss = 0
        label_masks = {}
        for sensor in self.graph.get_sensors(CandidateSensor):
            sensor(data)
            mask = data[sensor.fullname].clone().detach()
            label_masks[mask.shape] = mask
        for prop in self.graph.poi:
            # label (b, l)
            # pred  (b, l, c)
            # mask  (b, l)
            label, pred, mask = get_prop_result(prop, data)
            if self.inference_loss:
                # (b, l, c)
                _, pred_i, _ = get_prop_result(prop, data_i)
                # (b, l,)
                pred_i = pred_i.argmax(dim=-1)
            label = label.clone().detach()
            mask = mask.clone().float()
            label_mask = label_masks.get(mask.shape)
            if label_mask is not None:
                mask = mask * label_mask.float()
            #import pdb; pdb.set_trace()

            # class balance weighted
            num_token = float(prod(pred.shape[:-1]))
            num_classes = pred.shape[-1]
            balance_bias = num_token / self.balance_factor

            alpha = [((num_token + balance_bias) / ((label == class_index).sum().float() + balance_bias))
                     for class_index in range(num_classes)]

            if self.inference_loss:
                loss += structured_perceptron_exact_with_logits(
                    pred, label, mask, pred_i,
                    label_smoothing=self.label_smoothing,
                    gamma=self.focal_gamma,
                    alpha=alpha,
                    soft_penalty=self.soft_penalty
                )
            else:
                loss += sequence_cross_entropy_with_logits(
                    pred, label, mask,
                    label_smoothing=self.label_smoothing,
                    gamma=self.focal_gamma,
                    alpha=alpha
                )

        return loss



from torch import nn
import torch
import torch.nn.functional as F


class LSTMModel(nn.Module):

    def __init__(self, input_dim, hidden_dim, bidirectional=False,
                 num_layers=1, batch_size=1):
        super(LSTMModel, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.batch_size = batch_size
        # Define the LSTM layer
        self.lstm = nn.LSTM(input_size=self.input_dim, hidden_size=self.hidden_dim, num_layers=self.num_layers,
                            bidirectional=bidirectional)

    def init_hidden(self):
        # This is what we'll initialise our hidden state as
        return (torch.zeros(self.num_layers * 2, self.batch_size, self.hidden_dim),
                torch.zeros(self.num_layers * 2, self.batch_size, self.hidden_dim))

    def forward(self, input):
        # Forward pass through LSTM layer
        # shape of lstm_out: [input_size, batch_size, hidden_dim]
        # shape of self.hidden: (a, b), where a and b both
        # have shape (num_layers, batch_size, hidden_dim).
        lstm_out, share_hidden = self.lstm(input)

        return lstm_out


class PyTorchFC(nn.Module):
    def __init__(self, input_dim, output_dim):
        super(PyTorchFC, self).__init__()
        self.fc1 = nn.Linear(input_dim, output_dim)

    def forward(self, x):
        # x = F.relu(self.fc1(x))
        x = self.fc1(x[:, -1, :])
        return F.softmax(x, dim=1)


class PyTorchFCRelu(nn.Module):
    def __init__(self, input_dim, output_dim):
        super(PyTorchFCRelu, self).__init__()
        self.fc1 = nn.Linear(input_dim, output_dim)

    def forward(self, x):
        x = F.leaky_relu(self.fc1(x))
        # x = self.fc1(x[:, -1, :])
        return x


from itertools import cycle, repeat
from typing import Callable, List
from dataclasses import dataclass

from ..utils import consume, entuple
from .model.base import Mode
from .program import LearningBasedProgram


class BatchProgram(LearningBasedProgram):
    def __init__(self, *args, batch_size=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.batch_size = batch_size

    def train_epoch(self, dataset):
        # do not use super() because it call zero_grad for every step definitly
        self.model.mode(Mode.TRAIN)
        self.model.reset()
        self.opt.zero_grad()
        for index, data_item in enumerate(dataset):
            loss, metric, *output = self.model(data_item)
            if self.opt and loss:
                loss.backward()
                if index % self.batch_size == self.batch_size - 1:
                    self.opt.step()
                    self.opt.zero_grad()
            yield (loss, metric, *output[:1])
